"""Write a function that encrypts a string with a variable rotary cipher.

The function should take in a number and string and shift the string's
characters by that number:

>>> rot_encode(1, 'abcxyz')
'bcdyza'

It should be able to shift characters by any number:

>>> rot_encode(3, 'abcxyz')
'defabc'

It should preserve capitalization, whitespace, and any special characters:

>>> rot_encode(1, 'Wow! This is 100% amazing.')
'Xpx! Uijt jt 100% bnbajoh.'
"""


def rot_encode(shift, txt):
    """Encode `txt` by shifting its characters to the right."""

    # START SOLUTION

    encoded_txt = []
    for c in txt:
        if c.isalpha():
            # Convert `c` to its ascii value
            ascii_c = ord(c.lower())

            # Shift the ascii value. Then scale the resulting
            # number so it falls within a range of 0--26 (exclusive).
            # Then, mod 26 so values >= 26 loop back to 0. Add 97 to
            # turn the number back into valid ascii before converting
            # ascii number to a string
            encoded_c = chr(((ascii_c + shift - 97) % 26) + 97)
            if c.isupper():
                encoded_c = encoded_c.upper()

            encoded_txt.append(encoded_c)
        else:
            encoded_txt.append(c)
    return ''.join(encoded_txt)


def rot_encode2(shift, txt):
    """An alternative solution that doesn't involve manipulating ASCII."""

    alpha = list('abcdefghijklmnopqrstuvwxyz')
    shifted_alpha = []
    for i in range(0 + shift, len(alpha) + shift):
        shifted_alpha.append(alpha[(i % 26)])

    lookup_tbl = {}
    for i in range(len(alpha)):
        lookup_tbl[alpha[i]] = shifted_alpha[i]

    encoded_txt = []
    for c in txt:
        if c.isalpha():
            encoded_c = lookup_tbl[c.lower()]
            if c.isupper():
                encoded_c = encoded_c.upper()

            encoded_txt.append(encoded_c)
        else:
            encoded_txt.append(c)

    return ''.join(encoded_txt)


def rot_encode3(shift, txt):
    """Lastly, a cheeky solution that uses a circular list to shift chars."""

    alpha = list('abcdefghijklmnopqrstuvwxyz')
    # Make a circular list by appending the list to itself
    alpha.append(alpha)

    # This will help us loop through characters in the shifted alphabet
    def shifted_alpha(alpha):
        """Return iterator that yields the next char in shifted alphabet."""

        for c in alpha:
            if isinstance(c, list):
                yield from shifted_alpha(c)
            else:
                yield c

    # Create shifted alphabet by slicing `alpha` by `shift`. We'll
    # retrieve one char at a time from `sh_alpha` by calling
    # `next(sh_alpha)`
    sh_alpha = shifted_alpha(alpha[shift:])
    lookup_tbl = {}
    for c in alpha:
        lookup_tbl[c] = next(sh_alpha)

    # Use `lookup_tbl` to encode text
    encoded_txt = []
    for c in txt:
        if c.isalpha():
            encoded_c = lookup_tbl[c.lower()]
            if c.isupper():
                encoded_c = encoded_c.upper()

            encoded_txt.append(encoded_c)
        else:
            encoded_txt.append(c)

    return ''.join(encoded_txt)

    # END SOLUTION


if __name__ == '__main__':
    import doctest
    if doctest.testmod().failed == 0:
        print('\n✨ ALL TESTS PASSED!\n')
